package edu.washington.cs.radishsim;

import com.google.common.collect.HashMultiset;
import com.google.common.collect.Multiset;

/**
 * Manages the details of EpochOffsets, their allocation and garbage collection.
 */
public class EpochOffsetManager {

   /** tracks reference counts of each epoch offset for this cpu */
   private Multiset<EpochOffset> refCountOfEpochOffset = HashMultiset.create();

   /**
    * The max value a local epoch can have in the base+offset scheme. NB: the value 0 is reserved
    * for the "untouched" epoch, which indicates that no access has yet taken place.
    */
   private final EpochOffset     MAX_BASEOFFSET_LOCAL_EPOCH_OFFSET;

   /** The max number of epoch offsets that can be allocated (live) simultaneously. */
   private final long            MAX_LIVE_EPOCH_OFFSETS;

   /** Starts at 1 since 0 means "untouched" */
   private EpochOffset           epochOffset           = EpochOffset.INITIAL_EPOCH_OFFSET;

   /**
    * true if we are reference-counting clocks (epoch offset values), false if we are using the
    * simpler base+offset accounting scheme.
    */
   private final boolean         GC_CLOCKS;

   public EpochOffsetManager(long bitsPerLocalEpoch, boolean gcClocks) {
      MAX_LIVE_EPOCH_OFFSETS = (1 << bitsPerLocalEpoch) - 1/* value 0 is reserved */;
      MAX_BASEOFFSET_LOCAL_EPOCH_OFFSET = new EpochOffset(MAX_LIVE_EPOCH_OFFSETS);
      GC_CLOCKS = gcClocks;
   }

   /** Returns the current EpochOffset */
   public EpochOffset current() {
      return epochOffset;
   }

   /**
    * Returns whether a new EpochOffset can be allocated, or false if the max number of live
    * EpochOffsets has been exhausted.
    */
   public boolean canAllocate() {
      if ( GC_CLOCKS ) {
         return refCountOfEpochOffset.elementSet().size() < MAX_LIVE_EPOCH_OFFSETS;
      } else { // base + offset scheme
         return EpochOffset.Compare.lt(epochOffset, MAX_BASEOFFSET_LOCAL_EPOCH_OFFSET);
      }
   }

   /**
    * Increment the current EpochOffset. If clock GC is enabled, this EpochOffset has an initial
    * reference count of 0.
    * 
    * NB: with clock GC we aren't actually allocating anything (since the ref count is 0); if the id
    * is never used then it can represent the next epoch id instead.
    */
   public EpochOffset incrementCurrent() {
      assert canAllocate();

      EpochOffset newOffset = epochOffset.getIncrementedEpochOffset();

      if ( GC_CLOCKS ) {
         assert !refCountOfEpochOffset.contains(newOffset);
      }

      epochOffset = newOffset;
      return epochOffset;
   }

   /** Increment the reference count of the given EpochOffset */
   public void incrRefCount(final EpochOffset epochId) {
      assert GC_CLOCKS;
      // bump ref count
      refCountOfEpochOffset.add(epochId);
   }

   /** Decrement the reference count of the given EpochOffset */
   public void decrRefCount(final EpochOffset epochId) {
      assert GC_CLOCKS;

      int rc = refCountOfEpochOffset.count(epochId);
      assert rc >= 1;
      refCountOfEpochOffset.remove(epochId);
   }

   public boolean isAllocated(final EpochOffset eo) {
      assert GC_CLOCKS;
      return refCountOfEpochOffset.contains(eo);
   }

   /** Used for testing. */
   Multiset<EpochOffset> getRefCounts() {
      return this.refCountOfEpochOffset;
   }

   /**
    * Reset the state of this manager.
    * 
    * @return the EpochOffset immediately before the reset
    */
   public EpochOffset reset() {
      EpochOffset old = epochOffset;
      epochOffset = EpochOffset.INITIAL_EPOCH_OFFSET;
      if ( GC_CLOCKS ) {
         refCountOfEpochOffset.clear();
      }
      return old;
   }

   /** The max value an EpochOffset can have in the base+offset scheme */
   public EpochOffset MAX_BASEOFFSET_VALUE() {
      assert !GC_CLOCKS;
      return MAX_BASEOFFSET_LOCAL_EPOCH_OFFSET;
   }

}
