package edu.washington.cs.radishsim;

import java.io.IOException;
import java.io.Writer;
import java.util.LinkedList;
import java.util.List;
import java.util.StringTokenizer;

import edu.washington.cs.radishsim.Machine.LeafReadConfiguration;

import joptsimple.*;

public class Knobs {

   public static final OptionSpecBuilder                 Help;
   public static final OptionSpecBuilder                 NoXasserts;
   public static final OptionSpec<String>                StatsFile;
   public static final OptionSpec<String>                ToSimulatorFifo;
   public static final OptionSpecBuilder                 CommTest;

   public static final OptionSpec<String>                Scheme;
   public static final OptionSpec<String>                Workload;
   public static final OptionSpec<Integer>               Threads;
   public static final OptionSpec<String>                Input;

   public static final OptionSpec<Integer>               Cores;

   // caches
   public static final OptionSpec<Integer>               LineSize;
   public static final OptionSpec<Integer>               L1Size;
   public static final OptionSpec<Integer>               L1Assoc;
   public static final OptionSpecBuilder                 UseL2;
   public static final OptionSpec<Integer>               L2Size;
   public static final OptionSpec<Integer>               L2Assoc;
   public static final OptionSpecBuilder                 UseL3;
   public static final OptionSpec<Integer>               L3Size;
   public static final OptionSpec<Integer>               L3Assoc;

   // below is all radish stuff
   public static final OptionSpec<String>                SimulationMode;
   public static final OptionSpecBuilder                 UseWholeLineSwChecks;
   public static final OptionSpecBuilder                 IgnoreStackRefs;
   public static final OptionSpecBuilder                 GcClocks;
   public static final OptionSpecBuilder                 TypeSafetyOpt;
   public static final OptionSpecBuilder                 InfiniteEpochs;
   public static final OptionSpecBuilder                 RemoteAccessesAffectLRU;
   public static final OptionSpec<Integer>               SwCheckCost;
   public static final OptionSpec<Integer>               HwCheckCost;
   public static final OptionSpec<Integer>               EpochRolloverCost;
   public static final OptionSpecBuilder                 EvictUselessMetadata;
   public static final OptionSpec<LeafReadConfiguration> LeafReads;
   public static final OptionSpecBuilder                 MetadataLLCEvictionPolicy;
   public static final OptionSpec<Integer>               AsyncProcessingCores;
   public static final OptionSpecBuilder                 CompactMd;
   public static final OptionSpec<Integer>               CacheUsageSnapshotFrequency;

   public static final OptionParser                      parser;

   private Knobs() {}

   static {
      parser = new OptionParser();
      BooleanParameters = new LinkedList<OptionSpecBuilder>();
      StringParameters = new LinkedList<OptionSpec<String>>();
      IntegerParameters = new LinkedList<OptionSpec<Integer>>();
      EnumParameters = new LinkedList<OptionSpec<? extends Enum<?>>>();

      Help = parser.accepts("help", "print this help message");
      NoXasserts = parser.accepts("no-xasserts", "ensure that eXpensive assert checks are disabled");
      CommTest = parser.accepts("comm-test", "test communication between front-end and simulator");
      StatsFile = parser.accepts("stats-file", "stats file to generate").withRequiredArg()
                        .defaultsTo("radishsim-stats.py");
      ToSimulatorFifo = parser.accepts("tosim-fifo",
                                       "named fifo used to get events from the front-end")
                              .withRequiredArg();

      // these are used to tag the output file, but don't affect the
      // simulation at all
      Scheme = registerString(parser.accepts("scheme", "text describing this simulation setup")
                                    .withRequiredArg().defaultsTo("<none/>"));
      Workload = registerString(parser.accepts("workload", "workload being run").withRequiredArg()
                                      .defaultsTo("<none/>"));
      Input = registerString(parser.accepts("input", "input being used").withRequiredArg()
                                   .defaultsTo("<none/>"));
      Threads = registerInt(parser.accepts("threads", "number of threads in the workload")
                                  .withRequiredArg().ofType(Integer.class).defaultsTo(0));

      // processor parameters
      Cores = registerInt(parser.accepts("cores", "number of cores to simulate").withRequiredArg()
                                .ofType(Integer.class).defaultsTo(0));

      // cache parameters
      LineSize = registerInt(parser.accepts("line-size", "Line size for all caches")
                                   .withOptionalArg().ofType(Integer.class).defaultsTo(64));
      L1Size = registerInt(parser.accepts("l1-size", "Size (in bytes) of each private L1 cache")
                                 .withOptionalArg().ofType(Integer.class)
                                 .defaultsTo(1 << 15/* 32KB */));
      L1Assoc = registerInt(parser.accepts("l1-assoc", "Associativity of each private L1 cache")
                                  .withOptionalArg().ofType(Integer.class).defaultsTo(8));

      UseL2 = registerBool(parser.accepts("use-l2", "Model a private L2 for each core"));
      L2Size = registerInt(parser.accepts("l2-size", "Size (in bytes) of the private L2 cache")
                                 .withOptionalArg().ofType(Integer.class)
                                 .defaultsTo(1 << 18/* 256KB */));
      L2Assoc = registerInt(parser.accepts("l2-assoc", "Associativity of the private L2 cache")
                                  .withOptionalArg().ofType(Integer.class).defaultsTo(8));

      UseL3 = registerBool(parser.accepts("use-l3", "Model an L3 cache shared amongst all cores"));
      L3Size = registerInt(parser.accepts("l3-size", "Size (in bytes) of the shared L3 cache")
                                 .withOptionalArg().ofType(Integer.class)
                                 .defaultsTo(1 << 23/* 8MB */));
      L3Assoc = registerInt(parser.accepts("l3-assoc", "Associativity of the shared L3 cache")
                                  .withOptionalArg().ofType(Integer.class).defaultsTo(16));

      SimulationMode = registerString(parser.accepts("sim-mode", "Simulation mode to use (one of: radish, fasttrack, baseline).").withRequiredArg()
                                      .defaultsTo("<none/>"));

      // radish stuff
      UseWholeLineSwChecks = registerBool(parser.accepts("line-sw-checks",
                                                         "Software VC checks verify the whole line (otherwise: just the bytes of a single memory op)"));
      IgnoreStackRefs = registerBool(parser.accepts("ignore-stack",
                                                    "Assume stack references are thread-private."));
      GcClocks = registerBool(parser.accepts("gc-clocks",
                                             "Enable reference counting of clock values."));
      TypeSafetyOpt = registerBool(parser.accepts("type-safety",
                                                  "Enable type-safety optimizations."));
      InfiniteEpochs = registerBool(parser.accepts("inf-epochs",
                                                   "Assume epoch counters never roll over."));
      RemoteAccessesAffectLRU = registerBool(parser.accepts("remote-accesses-affect-lru",
                                                            "Remote accesses update LRU cache state."));

      SwCheckCost = registerInt(parser.accepts("sw-check-cost",
                                               "The fixed cost (in cycles) of a software VC check: cost of gathering VCs, etc.")
                                      .withRequiredArg().ofType(Integer.class).defaultsTo(100));
      HwCheckCost = registerInt(parser.accepts("hw-check-cost",
                                               "The cost (in cycles) of a hardware vc check.")
                                      .withRequiredArg().ofType(Integer.class).defaultsTo(10));
      EpochRolloverCost = registerInt(parser.accepts("epoch-rollover-cost",
                                                     "The cost (in cycles) of an epoch rollover cost")
                                            .withRequiredArg().ofType(Integer.class)
                                            .defaultsTo(100000));
      EvictUselessMetadata = registerBool(parser.accepts("evict-useless-md",
                                                         "Evict useless (empty/non-leaf read) metadata preferentially over other (write/leaf read) metadata."));

      LeafReads = registerEnum(parser.accepts("leaf-reads",
                                              "Eagerly update leaf-read info for free.")
                                     .withRequiredArg().ofType(LeafReadConfiguration.class)
                                     .defaultsTo(LeafReadConfiguration.LAZY_LEAF_READS));

      MetadataLLCEvictionPolicy = registerBool(parser.accepts("md-llc-eviction-policy",
                                                              "Prefer evicting data over metadata in the LLC."));

      AsyncProcessingCores = registerInt(parser.accepts("async-cores",
                                                        "Cores used for processing asynchronous tasks.")
                                               .withRequiredArg().ofType(Integer.class)
                                               .defaultsTo(1));

      CompactMd = registerBool(parser.accepts("compact-md",
                                              "Enable best-effort use of the compact metadata representation."));

      CacheUsageSnapshotFrequency = registerInt(parser.accepts("cache-usage-snapshot-freq",
                                                               "Take periodic cache usage snapshots. Default: no snapshots.")
                                                      .withRequiredArg().ofType(Integer.class)
                                                      .defaultsTo(0));
   }

   /*
    * Below is the stuff that automatically allows certain flags ("registered" ones) to appear in
    * the stats output, without any additional effort.
    */

   private static final List<OptionSpecBuilder>             BooleanParameters;
   private static final List<OptionSpec<String>>            StringParameters;
   private static final List<OptionSpec<Integer>>           IntegerParameters;
   private static final List<OptionSpec<? extends Enum<?>>> EnumParameters;

   private static OptionSpec<String> registerString(OptionSpec<String> o) {
      StringParameters.add(o);
      return o;
   }

   private static OptionSpec<Integer> registerInt(OptionSpec<Integer> o) {
      IntegerParameters.add(o);
      return o;
   }

   private static OptionSpecBuilder registerBool(OptionSpecBuilder o) {
      BooleanParameters.add(o);
      return o;
   }

   private static <E extends Enum<E>> OptionSpec<E> registerEnum(OptionSpec<E> o) {
      EnumParameters.add(o);
      return o;
   }

   private static String format(String flag) {
      // strip brackets
      String noBrackets = flag.replaceAll("\\[", "").replaceAll("\\]", "");

      // tokenize on dashes
      String result = "";
      StringTokenizer tok = new StringTokenizer(noBrackets, "-");
      while ( tok.hasMoreTokens() ) {
         String t = tok.nextToken();
         // capitalize each token
         result += (t.substring(0, 1).toUpperCase() + t.substring(1));
      }

      return result;
   }

   public static void dumpRegisteredParams(Writer w) throws IOException {
      for ( OptionSpecBuilder osb : BooleanParameters ) {
         String value = RadishSim.Options.has(osb) ? "True" : "False";
         w.write("'" + format(osb.toString()) + "': " + value + ", ");
      }
      for ( OptionSpec<String> os : StringParameters ) {
         w.write("'" + format(os.toString()) + "': '" + RadishSim.Options.valueOf(os) + "', ");
      }
      for ( OptionSpec<? extends Enum<?>> os : EnumParameters ) {
         w.write("'" + format(os.toString()) + "': '" + RadishSim.Options.valueOf(os).toString()
                 + "', ");
      }
      for ( OptionSpec<Integer> os : IntegerParameters ) {
         if ( os == Cores ) {
            w.write("'" + format(os.toString()) + "': '" + RadishSim.Options.valueOf(os) + "p', ");
         } else if ( os == L1Size || os == L2Size || os == L3Size ) {
            int kb = RadishSim.Options.valueOf(os) / 1024;
            w.write("'" + format(os.toString()) + "': '" + kb + "KB', ");
         } else {
            w.write("'" + format(os.toString()) + "': " + RadishSim.Options.valueOf(os) + ", ");
         }
      }
   }

}
