package edu.washington.cs.radishsim;

import static edu.washington.cs.radishsim.EventType.COMMUNICATION_TEST;

import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;

import edu.washington.cs.radishsim.Machine.SimulationMode;

import joptsimple.OptionSet;

public class RadishSim {

   /** enable checking of computationally expensive asserts */
   public static final boolean  XASSERTS         = true;

   public static final ThreadId INVALID_THREADID = new ThreadId(-1);

   static OptionSet             Options;

   private static int           maxLiveThreads;
   private static int           numSpawnedThreads;
   private static int           currentLiveThreads;
   private static long          insnsExecuted;
   private static long          stackAccesses;

   /**
    * @param args
    *           command line arguments
    * @throws IOException
    */
   public static void main(String[] args) throws IOException {

      Options = Knobs.parser.parse(args);

      if ( Options.has(Knobs.Help) ) {
         Knobs.parser.printHelpOn(System.out);
         return;
      }

      if ( Options.has(Knobs.NoXasserts) && RadishSim.XASSERTS ) { //
         throw new IllegalStateException("XASSERTS are enabled when they shouldn't be!");
      }

      final SimulationMode simMode;
      if ( Options.valueOf(Knobs.SimulationMode).equals("radish") ) {
         simMode = SimulationMode.RADISH;
      } else if ( Options.valueOf(Knobs.SimulationMode).equals("baseline") ) {
         simMode = SimulationMode.BASELINE;
      } else if ( Options.valueOf(Knobs.SimulationMode).equals("fasttrack") ) {
         simMode = SimulationMode.FASTTRACK;
      } else {
         throw new IllegalStateException("Invalid simulation mode: "
                                         + Options.valueOf(Knobs.SimulationMode));
      }

      DataInputStream in;
      try {
         FileInputStream fis = new FileInputStream(Options.valueOf(Knobs.ToSimulatorFifo));
         in = new DataInputStream(new BufferedInputStream(fis));
      } catch ( FileNotFoundException fnf ) {
         fnf.printStackTrace();
         return;
      }

      final long startTime = System.currentTimeMillis();

      MemorySystemConstants.setLineSize(Options.valueOf(Knobs.LineSize));

      Machine.MachineParams<RadishLine> p = new Machine.MachineParams<RadishLine>() {

         SimulationMode simulationMode() {
            return simMode;
         }

         int numProcessors() {
            return Options.valueOf(Knobs.Cores);
         }

         CacheConfiguration<RadishLine> l1config() {
            // NB: be lazy and return a new object each time; shouldn't affect anything since the
            // values are always the same
            return new CacheConfiguration<RadishLine>() {
               {
                  cacheSize = Options.valueOf(Knobs.L1Size);
                  lineSize = MemorySystemConstants.LINE_SIZE();
                  assoc = Options.valueOf(Knobs.L1Assoc);
                  level = CacheLevel.L1;
               }
            };
         }

         boolean useL2() {
            return Options.has(Knobs.UseL2);
         }

         CacheConfiguration<RadishLine> l2config() {
            return new CacheConfiguration<RadishLine>() {
               {
                  cacheSize = Options.valueOf(Knobs.L2Size);
                  lineSize = MemorySystemConstants.LINE_SIZE();
                  assoc = Options.valueOf(Knobs.L2Assoc);
                  level = CacheLevel.L2;
               }
            };
         }

         boolean useL3() {
            return Options.has(Knobs.UseL3);
         }

         CacheConfiguration<RadishLine> l3config() {
            return new CacheConfiguration<RadishLine>() {
               {
                  cacheSize = Options.valueOf(Knobs.L3Size);
                  lineSize = MemorySystemConstants.LINE_SIZE();
                  assoc = Options.valueOf(Knobs.L3Assoc);
                  level = CacheLevel.L3;
               }
            };
         }

         LineFactory<RadishLine> lineFactory() {
            return new LineFactory<RadishLine>() {
               @Override
               public RadishLine create() {
                  return new RadishLine(RadishLine.Compacted.COMPACT);
               }

               @Override
               public RadishLine create(LineAddress la, RadishLine.Compacted c) {
                  return RadishSim.createRadishLine(la, c);
               }
            };
         }

         int bitsPerLocalEpochOffset() {
            int mdBits = MemorySystemConstants.EXPANDED_METADATA_BYTES_PER_DATA_BYTE() * 8;
            int metadataBits = mdBits - 3 /* local perms + IHS */;
            int numBitsPerLocalEpoch = metadataBits / 2; // for each of r/w epochs
            if ( Options.has(Knobs.TypeSafetyOpt) ) numBitsPerLocalEpoch = 14;
            return Options.has(Knobs.InfiniteEpochs) ? 64 : numBitsPerLocalEpoch;
         }

         boolean useWholeLineSoftwareChecks() {
            return Options.has(Knobs.UseWholeLineSwChecks);
         }

         boolean useClockGC() {
            return Options.has(Knobs.GcClocks);
         }

         boolean useTypeSafetyOpt() {
            return Options.has(Knobs.TypeSafetyOpt);
         }

         boolean remoteAccessesAffectLRU() {
            return Options.has(Knobs.RemoteAccessesAffectLRU);
         }

         int swCheckCost() {
            return Options.valueOf(Knobs.SwCheckCost);
         }

         int hwCheckCost() {
            return Options.valueOf(Knobs.HwCheckCost);
         }

         int epochRolloverCost() {
            return Options.valueOf(Knobs.EpochRolloverCost);
         }

         boolean evictUselessMetadata() {
            return Options.has(Knobs.EvictUselessMetadata);
         }

         Machine.LeafReadConfiguration leafReads() {
            return Options.valueOf(Knobs.LeafReads);
         }

         boolean metadataLLCEvictionPolicy() {
            return Options.has(Knobs.MetadataLLCEvictionPolicy);
         }

         int asyncProcessingCores() {
            return Options.valueOf(Knobs.AsyncProcessingCores);
         }

         boolean simulateSoftwareChecks() {
            return true;
         }

         boolean compactMetadata() {
            return Options.has(Knobs.CompactMd);
         }

         int cacheUsageSnapshotFrequencyInCycles() {
            return Options.valueOf(Knobs.CacheUsageSnapshotFrequency);
         }
      };

      Machine<RadishLine> sim = new Machine<RadishLine>(p);

      System.err.println("[radishsim] starting simulation...");

      //long eventsProcessed = 0;
      while ( true ) {
         try {
            final Event e = getNextEvent(in);
            boolean simulationFinished = handleEvent(e, sim);
            //eventsProcessed++;
            if ( simulationFinished ) break;

         } catch ( EOFException eof ) {
            break;
         }
      }

      in.close();
      double hours = (System.currentTimeMillis() - startTime) / (double)(1000 * 60 * 60);
      generateStats(hours, sim);

   } // end main()

   public static RadishLine createRadishLine(LineAddress la, RadishLine.Compacted c) {

      if ( la instanceof MetadataLineAddress ) {
         switch ( c ) {
         case COMPACT:
            if ( !MemorySystemConstants.COMPACT_MD_ALIGNED(la.get()) ) {
               // ensure compact md lines are aligned
               MetadataLineAddress mla = (MetadataLineAddress) la;
               long aligned = mla.get() & MemorySystemConstants.COMPACT_MD_ALIGNMENT_MASK();
               MetadataLineAddress alignedMla = new MetadataLineAddress(aligned, mla.owner());
               return new RadishLine(RadishLine.Compacted.COMPACT, alignedMla);

            } else { // md that's already aligned
               return new RadishLine(RadishLine.Compacted.COMPACT, la);
            }
         case EXPANDED:
            return new RadishLine(RadishLine.Compacted.EXPANDED, la);
         default:
            throw new IllegalArgumentException(String.valueOf(c));
         }

      } else { // data line; doesn't matter whether it's compact or expanded
         return new RadishLine(RadishLine.Compacted.EXPANDED, la);
      }
   }

   private static Event getNextEvent(DataInputStream in) throws IOException {
      Event e = new Event(EventType.fromInt(in.readByte()), in.readShort());
      e.addr = in.readLong();
      e.memOpSize = in.readInt();
      byte bits = in.readByte();
      e.stackRef = (bits & 0x1) == 1;
      e.isLifeLock = (bits & 0x2) == 2;
      e.syncObject = in.readLong();
      e.hbSourceTid = new ThreadId(in.readShort());
      e.insnCount = in.readByte();

      /* for communication testing, all events should be comm tests */
      if ( Options.has(Knobs.CommTest) ) {
         assert e.type == COMMUNICATION_TEST;
         assert e.tid.get() == 10;
         assert e.addr == 11;
         assert e.memOpSize == 12;
         assert e.stackRef && e.isLifeLock;
         assert e.syncObject == 14;
         assert e.hbSourceTid.get() == 15;
         assert e.insnCount == 16;
      }

      return e;
   }

   /**
    * Dispatch the given event to the simulator code.
    * 
    * @param e
    *           the next event from the front-end
    * @return true when the simulation is finished, false otherwise
    */
   private static boolean handleEvent(final Event e, Machine<RadishLine> machine) {
      CpuId cpuid = machine.cpuOfTid(e.tid);

      switch ( e.type ) {

      case ROI_START:
      case ROI_FINISH:
      case MEMORY_ALLOCATION:
      case MEMORY_FREE:
      case THREAD_BLOCKED:
      case THREAD_UNBLOCKED:
         // NOPs for now
         break;

      case THREAD_START:
         currentLiveThreads++;
         numSpawnedThreads++;
         maxLiveThreads = Math.max(maxLiveThreads, currentLiveThreads);
         break;

      case THREAD_FINISH:
         currentLiveThreads--;
         // when main thread exits, tear down simulation
         if ( e.tid.get() == 0 ) return true;
         break;

      case HAPPENS_BEFORE_SOURCE:
         machine.syncOp(cpuid, SynchronizationOperationType.SOURCE, false, e.syncObject);
         break;
      case HAPPENS_BEFORE_SINK:
         machine.syncOp(cpuid, SynchronizationOperationType.SINK,
                        !e.hbSourceTid.equals(INVALID_THREADID), e.syncObject);
         break;

      case MEMORY_READ:
         machine.cacheRead(cpuid, e.addr, e.memOpSize, doMetadataAccess(e));
         break;
      case MEMORY_WRITE:
         machine.cacheWrite(cpuid, e.addr, e.memOpSize, doMetadataAccess(e));
         break;

      case BASIC_BLOCK:
         insnsExecuted += e.insnCount;
         machine.insnsExecuted(cpuid, e.insnCount);
         break;

      case COMMUNICATION_TEST:
         assert Options.has(Knobs.CommTest);
         break;

      case INVALID_EVENT:
         System.out.println("Invalid event found: ");
         System.out.println(e);
         break;

      default:
         assert false : "Impossible event type";
      }

      return false; // not done processing events yet

   } // end handleEvent()

   /**
    * Whether to perform a metadata access for this memory event or not.
    */
   private static boolean doMetadataAccess(final Event e) {
      // are we ignoring stack accesses?
      if ( e.stackRef && Options.has(Knobs.IgnoreStackRefs) ) {
         stackAccesses++;
         return false;
      }

      // do the metadata access
      return true;
   }

   private static void generateStats(double simRuntimeHours, Machine<RadishLine> machine)
         throws IOException {
      System.err.println("[radishsim] exiting...");

      // each stat is dumped as a Python dictionary object

      StringWriter prefix = new StringWriter();
      prefix.write("{'RadishStat':True, ");
      Knobs.dumpRegisteredParams(prefix);

      String suffix = "}" + System.getProperty("line.separator");

      // check for filename collisions and rename around them
      String statsFilename = Options.valueOf(Knobs.StatsFile);
      File f = new File(statsFilename);
      while ( f.exists() ) {
         statsFilename += ".1";
         f = new File(statsFilename);
      }
      BufferedWriter statsFd = new BufferedWriter(new FileWriter(f));

      // dump stats from the caches
      machine.dumpStats(statsFd, prefix.toString(), suffix);

      // dump "global" stats
      
      statsFd.write(prefix.toString() + "'SimulationRunningTimeHours': "
                    + String.format("%.2f", simRuntimeHours) + suffix);
      
      double gigs = Runtime.getRuntime().totalMemory() / (double) (1 << 30);
      String memUsage = "'MemUsageGB': " + String.format("%.2f", gigs);
      statsFd.write(prefix.toString() + memUsage + suffix);
      
      statsFd.write(prefix.toString() + "'MaxLiveThreads': " + maxLiveThreads + suffix);
      statsFd.write(prefix.toString() + "'NumSpawnedThreads': " + numSpawnedThreads + suffix);
      statsFd.write(prefix.toString() + "'g_StackAccesses': " + stackAccesses + suffix);
      statsFd.write(prefix.toString() + "'g_Instructions': " + insnsExecuted + suffix);

      statsFd.close();

      System.err.println("[radishsim] finished");
   }

}
